@@ -4,6 +4,7 @@
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE MagicHash                  #-}
 {-# LANGUAGE RankNTypes                 #-}
+{-# LANGUAGE TemplateHaskellQuotes      #-}
 {-# LANGUAGE TypeFamilies               #-}
 {-# LANGUAGE UnboxedTuples              #-}
 {-# LANGUAGE UnliftedFFITypes           #-}
@@ -95,11 +96,18 @@ module Data.Text.Short.Internal
     , BS.ByteString
     , T.Text
     , module Prelude
+    , module Data.Monoid
 
       -- ** Internals
     , isValidUtf8
     ) where
 
+import           Prelude
+  (Bool(..), Ordering(..), Int, Char, String, Maybe(..), IO, Eq, Ord, Num, Read,
+   Show, ($), ($!), (.), (==), (/=), (+), (*), (-), (>>), (<=), (<), (>), (>=),
+   compare, show, showsPrec, readsPrec, abs, return, fmap, error,
+   undefined, otherwise, fromIntegral, max, min, not, fst, snd, map, seq, fail, maybe)
+
 import           Control.DeepSeq                (NFData)
 import           Control.Monad.ST               (stToIO)
 import           Data.Binary
@@ -117,10 +125,10 @@ import           Data.Hashable                  (Hashable)
 import           Data.Typeable                  (Typeable)
 import qualified Data.List                      as List
 import           Data.Maybe                     (fromMaybe, isNothing)
+import           Data.Monoid                    (Monoid, mempty, mconcat)
 import           Data.Semigroup
 import qualified Data.String                    as S
 import qualified Data.Text                      as T
-import qualified Data.Text.Encoding             as T
 import           Foreign.C
 import           GHC.Base                       (assert, unsafeChr)
 import qualified GHC.CString                    as GHC
@@ -131,16 +139,19 @@ import qualified GHC.Exts
 import qualified GHC.Foreign                    as GHC
 import           GHC.IO.Encoding
 import           GHC.ST
-import           Prelude                        hiding (all, any, break, concat,
-                                                 drop, dropWhile, filter, foldl,
-                                                 foldl1, foldr, foldr1, head,
-                                                 init, last, length, null,
-                                                 replicate, reverse, span,
-                                                 splitAt, tail, take, takeWhile)
 import           System.IO.Unsafe
 import           Text.Printf                    (PrintfArg, formatArg,
                                                  formatString)
 
+import qualified Language.Haskell.TH.Syntax     as TH
+
+#if MIN_VERSION_text(2,0,0)
+import qualified Data.Text.Internal             as TI
+import qualified Data.Text.Array                as TA
+#else
+import qualified Data.Text.Encoding             as T
+#endif
+
 import qualified PrimOps
 
 -- | A compact representation of Unicode strings.
@@ -151,10 +162,11 @@ import qualified PrimOps
 --
 -- This type relates to 'T.Text' as 'ShortByteString' relates to 'BS.ByteString' by providing a more compact type. Please consult the documentation of "Data.ByteString.Short" for more information.
 --
--- Currently, a boxed unshared 'T.Text' has a memory footprint of 6 words (i.e. 48 bytes on 64-bit systems) plus 2 or 4 bytes per code-point (due to the internal UTF-16 representation). Each 'T.Text' value which can share its payload with another 'T.Text' requires only 4 words additionally. Unlike 'BS.ByteString', 'T.Text' use unpinned memory.
+-- Currently, a boxed unshared 'T.Text' has a memory footprint of 6 words (i.e. 48 bytes on 64-bit systems) plus 1, 2, 3 or 4 bytes per code-point for text-2 (due to the internal UTF-8 representation) or 2 or 4 bytes per  code-point for text-1 (due to the internal UTF-16 representation). Each 'T.Text' value which can share its payload with another 'T.Text' requires only 4 words additionally. Unlike 'BS.ByteString', 'T.Text' use unpinned memory.
 --
 -- In comparison, the footprint of a boxed 'ShortText' is only 4 words (i.e. 32 bytes on 64-bit systems) plus 1, 2, 3, or 4 bytes per code-point (due to the internal UTF-8 representation).
--- It can be shown that for realistic data <http://utf8everywhere.org/#asian UTF-16 has a space overhead of 50% over UTF-8>.
+--
+-- It can be shown that for realistic data <http://utf8everywhere.org/#asian UTF-16, which is used by text-1, has a space overhead of 50% over UTF-8>.
 --
 -- __NOTE__: The `Typeable` instance isn't defined for GHC 7.8 (and older) prior to @text-short-0.1.3@
 --
@@ -221,7 +233,6 @@ instance PrintfArg ShortText where
   formatArg txt = formatString $ toString txt
 
 -- | The 'Binary' encoding matches the one for 'T.Text'
-#if MIN_VERSION_binary(0,8,1)
 instance Binary ShortText where
     put = put . toShortByteString
     get = do
@@ -229,15 +240,17 @@ instance Binary ShortText where
         case fromShortByteString sbs of
           Nothing -> fail "Binary.get(ShortText): Invalid UTF-8 stream"
           Just st -> return st
-#else
--- fallback via 'ByteString' instance
-instance Binary ShortText where
-    put = put . toByteString
-    get = do
-        bs <- get
-        case fromByteString bs of
-          Nothing -> fail "Binary.get(ShortText): Invalid UTF-8 stream"
-          Just st -> return st
+
+-- | Since 0.1.3
+instance TH.Lift ShortText where
+    -- TODO: Use DeriveLift with bytestring-0.11.2.0
+    lift t = [| fromString s |]
+      where s = toString t
+
+#if MIN_VERSION_template_haskell(2,17,0)
+    liftTyped = TH.unsafeCodeCoerce . TH.lift
+#elif MIN_VERSION_template_haskell(2,16,0)
+    liftTyped = TH.unsafeTExpCoerce . TH.lift
 #endif
 
 -- | \(\mathcal{O}(1)\) Test whether a 'ShortText' is empty.
@@ -633,12 +646,16 @@ foldr1 f st
 --
 -- prop> (toText . fromText) t == t
 --
--- This is currently not \(\mathcal{O}(1)\) because currently 'T.Text' uses UTF-16 as its internal representation.
--- In the event that 'T.Text' will change its internal representation to UTF-8 this operation will become \(\mathcal{O}(1)\).
+-- This is \(\mathcal{O}(1)\) with @text-2@.
+-- Previously it wasn't because 'T.Text' used UTF-16 as its internal representation.
 --
 -- @since 0.1
 toText :: ShortText -> T.Text
+#if MIN_VERSION_text(2,0,0)
+toText (ShortText (BSSI.SBS ba)) = TI.Text (TA.ByteArray ba) 0 (I# (GHC.Exts.sizeofByteArray# ba))
+#else
 toText = T.decodeUtf8 . toByteString
+#endif
 
 ----
 
@@ -657,21 +674,37 @@ toText = T.decodeUtf8 . toByteString
 --
 -- @since 0.1
 fromString :: String -> ShortText
-fromString []  = mempty
-fromString [c] = singleton c
-fromString s = ShortText . encodeStringShort utf8 . map r $ s
+fromString s = case s of
+  []  -> mempty
+  [c] -> singleton $ r c
+  _   -> ShortText . encodeStringShort utf8 . map r $ s
   where
     r c | isSurr (ord c) = '\xFFFD'
         | otherwise      = c
 
 -- | \(\mathcal{O}(n)\) Construct 'ShortText' from 'T.Text'
 --
--- This is currently not \(\mathcal{O}(1)\) because currently 'T.Text' uses UTF-16 as its internal representation.
--- In the event that 'T.Text' will change its internal representation to UTF-8 this operation will become \(\mathcal{O}(1)\).
+-- This is \(\mathcal{O}(1)\) with @text-2@ when the 'T.Text' is not sliced.
+-- Previously it wasn't because 'T.Text' used UTF-16 as its internal representation.
 --
 -- @since 0.1
 fromText :: T.Text -> ShortText
+#if MIN_VERSION_text(2,0,0)
+fromText (TI.Text (TA.ByteArray ba) off len) =
+    ShortText (BSSI.SBS (case sliceByteArray (TA.ByteArray ba) off len of TA.ByteArray ba' -> ba'))
+
+sliceByteArray :: TA.Array -> Int -> Int -> TA.Array
+sliceByteArray ta@(TA.ByteArray ba) 0 len
+    | len == I# (GHC.Exts.sizeofByteArray# ba)
+    = ta
+sliceByteArray ta off len = TA.run $ do
+    ma <- TA.new len
+    TA.copyI len ma 0 ta off
+    return ma
+
+#else
 fromText = fromByteStringUnsafe . T.encodeUtf8
+#endif
 
 -- | \(\mathcal{O}(n)\) Construct 'ShortText' from UTF-8 encoded 'ShortByteString'
 --
@@ -1271,7 +1304,12 @@ newByteArray (B (I# n#))
 {-# INLINE writeWord8Array #-}
 writeWord8Array :: MBA s -> B -> Word -> ST s ()
 writeWord8Array (MBA# mba#) (B (I# i#)) (W# w#)
-  = ST $ \s -> case GHC.Exts.writeWord8Array# mba# i# w# s of
+  = ST $ \s ->
+#if __GLASGOW_HASKELL__ >= 902
+      case GHC.Exts.writeWord8Array# mba# i# (GHC.Exts.wordToWord8# w#) s of
+#else
+      case GHC.Exts.writeWord8Array# mba# i# w# s of
+#endif
                  s' -> (# s', () #)
 {- not needed yet
 {-# INLINE indexWord8Array #-}
